# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Helpers for working with models in CLI scripts."""
from __future__ import annotations

__all__ = ["ModelCLI", "enumerate_known_models", "loadcli"]

import collections.abc as cabc
import importlib.abc
import importlib.resources as imr
import json
import os
import typing as t

import capellambse

try:
    import click

    class ModelCLI(click.ParamType):
        """Declare an option that loads a model.

        Use instances of this class for the ``type=`` argument to
        ``@click.option`` etc.

        Examples
        --------
        .. code-block:: python

           @click.command()
           @click.option("-m", "--model", type=capellambse.ModelCLI())
           def main(model: capellambse.MelodyModel) -> None:
               ...
        """

        name = "CAPELLA_MODEL"

        def convert(self, value: t.Any, param, ctx) -> capellambse.MelodyModel:
            if isinstance(value, capellambse.MelodyModel):
                return value

            try:
                return loadcli(value)
            except ValueError as err:
                self.fail(err.args[0], param, ctx)
                assert False

except ImportError:

    def ModelCLI(*__, **_):  # type: ignore[no-redef]
        """Raise a dependency error."""
        raise RuntimeError("click is not installed")


def enumerate_known_models() -> cabc.Iterator[importlib.abc.Traversable]:
    """Enumerate the models that are found in the ``known_models`` folder.

    In order to make a custom model known, place a JSON file in the
    ``known_models`` folder below the  installed ``capellambse`` package.
    It should contain a dictionary with the keyword arguments to
    :class:`capellambse.MelodyModel` - specifically it needs a ``path``,
    an ``entrypoint``, and any additional arguments that the underlying
    :class:`capellambse.filehandler.FileHandler` might need to gain
    access to the model.

    Be aware that relative paths in the JSON will be interpreted
    relative to the current working directory.
    """
    for i in imr.files(capellambse).joinpath("known_models").iterdir():
        if i.name.endswith(".json") and i.is_file():
            yield i


def loadcli(value: str | os.PathLike[str]) -> capellambse.MelodyModel:
    """Load a model from a file or JSON string.

    This function tries to load a model from the following inputs:

    - A str or PathLike pointing to an ``.aird`` file
    - A str or PathLike pointing to a ``.json`` file, which contains the
      arguments to instantiate a :class:`capellambse.MelodyModel`
    - The contents of such a JSON file (as string)

    See Also
    --------
    capellambse.cli_helpers.CLIModel :
        This function wrapped as a ``click.ParamType``
    """
    if isinstance(value, str):
        if value.endswith((".aird", ".json")):
            return _load_from_file(value)
        try:
            data = json.loads(value)
        except json.JSONDecodeError:
            return _load_from_file(value)
        return capellambse.MelodyModel(**data)

    if isinstance(value, os.PathLike):
        value = os.fspath(value)
        if isinstance(value, str):
            return _load_from_file(value)

    raise TypeError("value must be a str or PathLike returning str")


def _load_from_file(value: str) -> capellambse.MelodyModel:
    if value.endswith(".aird"):
        return capellambse.MelodyModel(value)

    if value.endswith(".json"):
        with open(value, encoding="utf-8") as file:
            data = json.load(file)
    else:
        value = value + ".json"
        for known in enumerate_known_models():
            if value == known.name:
                data = json.loads(known.read_text(encoding="utf-8"))
                break
        else:
            try:
                data = json.loads(value)
            except json.JSONDecodeError:
                raise ValueError(
                    "value is not a known model nor contains valid JSON"
                ) from None
    return capellambse.MelodyModel(**data)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import collections.abc as cabc
import contextlib
import datetime
import sys
import typing as t

import markupsafe
import pytest

import capellambse
from capellambse.extensions import reqif
from capellambse.model import common

# pylint: disable=pointless-statement, redefined-outer-name


def test_AttributeAuditor_tracks_uuids_of_accessed_ModelElements(
    model: capellambse.MelodyModel,
):
    comp, comp1, comp2 = model.la.all_components[:3]
    expected = {comp.uuid, comp2.uuid}
    with capellambse.AttributeAuditor(
        model, {"name", "description"}
    ) as recorded_ids:
        comp.name
        comp1.uuid
        comp2.description

    assert expected == recorded_ids


def test_AttributeAuditor_handles_recursive_uuid(
    model: capellambse.MelodyModel,
):
    comp = model.la.all_components[0]
    with capellambse.AttributeAuditor(model, {"uuid"}) as recorded_ids:
        comp.uuid

    assert {comp.uuid} == recorded_ids


def test_AttributeAuditor_destroys_model_reference_when_detach(
    model: capellambse.MelodyModel,
):
    auditor = capellambse.AttributeAuditor(model, {"name"})

    auditor.detach()

    assert auditor.model is None


@pytest.fixture
def audit_events() -> cabc.Iterator[list[tuple[t.Any, ...]]]:
    events: list[tuple[t.Any, ...]] | None = []

    def audit(event: str, args: tuple[t.Any, ...]) -> None:
        nonlocal events
        if events is None or not event.startswith("capellambse."):
            return
        events.append((event, *args))

    sys.addaudithook(audit)
    try:
        assert events is not None
        yield events
    finally:
        events = None


@contextlib.contextmanager
def prohibit_events(*events: str) -> cabc.Iterator[None]:
    active = True
    events = set(events)  # type: ignore

    def audit(event: str, _: tuple[t.Any, ...]) -> None:
        nonlocal active
        if active:
            assert event not in events

    sys.addaudithook(audit)
    try:
        yield
    finally:
        active = False


@pytest.mark.parametrize(
    ["obj_id", "attr", "accessor_type"],
    [
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "name",
            common.AttributeProperty,
            id="AttributeProperty",
        ),
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "is_abstract",
            common.BooleanAttributeProperty,
            id="BooleanAttributeProperty",
        ),
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "description",
            common.HTMLAttributeProperty,
            id="HTMLAttributeProperty",
        ),
        pytest.param(
            "303c2a4d-0eff-41e6-b7e8-0e500cfa38f7",
            "value",
            common.NumericAttributeProperty,
            id="NumericAttributeProperty",
        ),
        pytest.param(
            "b97c09b5-948a-46e8-a656-69d764ddce7d",
            "value",
            common.DatetimeAttributeProperty,
            id="DatetimeAttributeProperty",
        ),
        pytest.param(
            "1ca7b206-be29-4315-a036-0b532b26a191",
            "type",
            common.EnumAttributeProperty,
            id="EnumAttributeProperty",
        ),
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "components",
            common.DirectProxyAccessor,
            id="DirectProxyAccessor",
        ),
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "allocated_functions",
            common.LinkAccessor,
            id="LinkAccessor",
        ),
        pytest.param(
            "0ee1fd1c-db9c-4fa6-8789-197f623b96c0",
            "type",
            common.AttrProxyAccessor,
            id="AttrProxyAccessor",
        ),
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "parent",
            common.ParentAccessor,
            id="ParentAccessor",
        ),
        pytest.param(
            "7abe3167-31d9-4c89-9aad-ef422e9deb6b",
            "specification",
            common.SpecificationAccessor,
            id="SpecificationAccessor",
        ),
        pytest.param(
            "0ee1fd1c-db9c-4fa6-8789-197f623b96c0",
            "min_card",
            common.RoleTagAccessor,
            id="RoleTagAccessor",
        ),
    ],
)
def test_attribute_access_fires_exactly_one_getattr_event(
    model: capellambse.MelodyModel,
    audit_events: list[tuple[t.Any, ...]],
    obj_id: str,
    attr: str,
    accessor_type: type[t.Any],
) -> None:
    obj = model.by_uuid(obj_id)
    descriptor = getattr(type(obj), attr, None)
    assert descriptor is not None, f"{type(obj).__name__} has no {attr}"
    assert isinstance(descriptor, accessor_type), "Bad descriptor type"
    audit_events.clear()

    getattr(obj, attr)

    events = [i for i in audit_events if i[0] == "capellambse.getattr"]
    assert len(events) == 1
    event = events[0]
    _, ev_obj, ev_attr, _ = event
    assert ev_obj == obj
    assert ev_attr == attr


@pytest.mark.parametrize(
    ["obj_id", "attr", "accessor_type"],
    [
        pytest.param(
            "3c2d312c-37c9-41b5-8c32-67578fa52dc3",
            "relations",
            reqif.RequirementsRelationAccessor,
            id="RequirementsRelationAccessor",
        ),
        pytest.param(
            "3c2d312c-37c9-41b5-8c32-67578fa52dc3",
            "related",
            reqif.ElementRelationAccessor,
            id="ElementRelationAccessor",
        ),
    ],
)
def test_attribute_access_fires_read_attribute_events(
    model: capellambse.MelodyModel,
    audit_events: list[tuple[t.Any, ...]],
    obj_id: str,
    attr: str,
    accessor_type: type[t.Any],
) -> None:
    """The last event is the one assigned to the high-level call."""

    obj = model.by_uuid(obj_id)
    descriptor = getattr(type(obj), attr, None)
    assert descriptor is not None, f"{type(obj).__name__} has no {attr}"
    assert isinstance(descriptor, accessor_type), "Bad descriptor type"
    audit_events.clear()

    getattr(obj, attr)

    events = [i for i in audit_events if i[0] == "capellambse.getattr"]
    assert len(events) > 1
    event = events[-1]
    _, ev_obj, ev_attr, _ = event
    assert ev_obj == obj
    assert ev_attr == attr


@pytest.mark.parametrize(
    ["obj_id", "attr", "value_factory", "accessor_type"],
    [
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "name",
            lambda _: "New name",
            common.AttributeProperty,
            id="AttributeProperty",
        ),
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "is_abstract",
            lambda o: not o.is_abstract,
            common.BooleanAttributeProperty,
            id="BooleanAttributeProperty",
        ),
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "description",
            lambda _: markupsafe.Markup("<h1>Thing</h1>"),
            common.HTMLAttributeProperty,
            id="HTMLAttributeProperty",
        ),
        pytest.param(
            "303c2a4d-0eff-41e6-b7e8-0e500cfa38f7",
            "value",
            lambda o: o.value + 1,
            common.NumericAttributeProperty,
            id="NumericAttributeProperty",
        ),
        pytest.param(
            "b97c09b5-948a-46e8-a656-69d764ddce7d",
            "value",
            lambda _: datetime.datetime(2019, 7, 23, 17, 45, 30),
            common.DatetimeAttributeProperty,
            id="DatetimeAttributeProperty",
        ),
        pytest.param(
            "1ca7b206-be29-4315-a036-0b532b26a191",
            "type",
            lambda _: "SHARED_DATA",
            common.EnumAttributeProperty,
            id="EnumAttributeProperty",
        ),
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "components",
            lambda _: [],
            common.DirectProxyAccessor,
            id="DirectProxyAccessor",
        ),
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "allocated_functions",
            lambda _: [],
            common.LinkAccessor,
            id="LinkAccessor",
        ),
        pytest.param(
            "0ee1fd1c-db9c-4fa6-8789-197f623b96c0",
            "type",
            lambda o: o._model.by_uuid("bbc296e1-ed4c-40cf-b37d-c8eb8613228a"),
            common.AttrProxyAccessor,
            id="AttrProxyAccessor",
        ),
        pytest.param(
            "3c2d312c-37c9-41b5-8c32-67578fa52dc3",
            "relations",
            lambda o: [],
            reqif.RequirementsRelationAccessor,
            id="RequirementsRelationAccessor",
        ),
    ],
)
def test_attribute_assignment_fires_exactly_one_setattr_event(
    model: capellambse.MelodyModel,
    audit_events: list[tuple[t.Any, ...]],
    obj_id: str,
    attr: str,
    value_factory: cabc.Callable[[capellambse.ModelObject], t.Any],
    accessor_type: type[t.Any],
) -> None:
    obj = model.by_uuid(obj_id)
    descriptor = getattr(type(obj), attr, None)
    assert descriptor is not None, f"{type(obj).__name__} has no {attr}"
    assert isinstance(descriptor, accessor_type), "Bad descriptor type"
    value = value_factory(obj)
    audit_events.clear()

    setattr(obj, attr, value)

    events = [i for i in audit_events if i[0] == "capellambse.setattr"]
    assert len(events) == 1
    _, actual_obj, actual_attr, actual_value = events[0]
    assert actual_obj is obj
    assert actual_attr == attr
    assert actual_value == value


@pytest.mark.parametrize(
    ["obj_id", "attr", "args_factory", "accessor_type"],
    [
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "components",
            lambda m: {
                "name": "Unfair advantages",
                "allocated_functions": [
                    m.by_uuid("c1a42acc-1f53-42bb-8404-77a5c08c414b")
                ],
            },
            common.DirectProxyAccessor,
            id="DirectProxyAccessor",
        ),
        pytest.param(
            "85d41db2-9e17-438b-95cf-49342452ddf3",
            "relations",
            lambda m: {
                "target": m.by_uuid("4c1f2b5d-0641-42c7-911f-7a42928580b8"),
            },
            reqif.RequirementsRelationAccessor,
            id="RequirementsRelationAccessor",
        ),
    ],
)
def test_creating_objects_fires_exactly_one_create_event(
    model: capellambse.MelodyModel,
    audit_events: list[tuple[t.Any, ...]],
    obj_id: str,
    attr: str,
    args_factory: cabc.Callable[[capellambse.MelodyModel], dict[str, t.Any]],
    accessor_type: type,
) -> None:
    obj = model.by_uuid(obj_id)
    descriptor = getattr(type(obj), attr, None)
    assert descriptor is not None, f"{type(obj).__name__} has no {attr}"
    assert isinstance(descriptor, accessor_type), "Bad descriptor type"
    create_args = args_factory(model)
    target = getattr(obj, attr)
    audit_events.clear()

    with prohibit_events("capellambse.insert", "capellambse.setattr"):
        new_obj = target.create(**create_args)

    event_filter = {
        "capellambse.create",
        "capellambse.insert",
        "capellambse.setattr",
    }
    events = [i for i in audit_events if i[0] in event_filter]
    eventnames = [i[0] for i in events]
    assert eventnames == ["capellambse.create"]
    _, *ev = events[0]
    assert ev[0] == obj
    assert ev[1] == attr
    assert ev[2] is new_obj


@pytest.mark.parametrize(
    ["obj_id", "attr", "accessor_type"],
    [
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "components",
            common.DirectProxyAccessor,
            id="DirectProxyAccessor",
        ),
        pytest.param(
            "90517d41-da3e-430c-b0a9-e3badf416509",
            "exchanges",
            common.LinkAccessor,
            id="LinkAccessor",
        ),
        pytest.param(
            "0c7b7899-49a7-4e41-ab11-eb7d9c2becf6",
            "do_activity",
            common.AttrProxyAccessor,
            id="AttrProxyAccessor",
        ),
        pytest.param(
            "85d41db2-9e17-438b-95cf-49342452ddf3",
            "relations",
            reqif.RequirementsRelationAccessor,
            id="RequirementsRelationAccessor",
        ),
    ],
)
def test_deleting_an_object_from_a_list_fires_one_delete_event(
    model: capellambse.MelodyModel,
    audit_events: list[tuple[t.Any, ...]],
    obj_id: str,
    attr: str,
    accessor_type: type,
) -> None:
    obj = model.by_uuid(obj_id)
    descriptor = getattr(type(obj), attr, None)
    assert descriptor is not None, f"{type(obj).__name__} has no {attr}"
    assert isinstance(descriptor, accessor_type), "Bad descriptor type"
    audit_events.clear()
    target = getattr(obj, attr)

    del target[0]

    events = [i for i in audit_events if i[0] == "capellambse.delete"]
    assert len(events) == 1
    _, *ev = events[0]
    assert ev[0] is obj
    assert ev[1] == attr
    assert ev[2] == 0


@pytest.mark.parametrize(
    ["obj_id", "attr", "accessor_type"],
    [
        pytest.param(
            "0d2edb8f-fa34-4e73-89ec-fb9a63001440",
            "components",
            common.DirectProxyAccessor,
            id="DirectProxyAccessor",
        ),
        pytest.param(
            "90517d41-da3e-430c-b0a9-e3badf416509",
            "exchanges",
            common.LinkAccessor,
            id="LinkAccessor",
        ),
        pytest.param(
            "0c7b7899-49a7-4e41-ab11-eb7d9c2becf6",
            "do_activity",
            common.AttrProxyAccessor,
            id="AttrProxyAccessor",
        ),
        pytest.param(
            "85d41db2-9e17-438b-95cf-49342452ddf3",
            "relations",
            reqif.RequirementsRelationAccessor,
            id="RequirementsRelationAccessor",
        ),
    ],
)
def test_deleting_an_entire_list_attribute_fires_one_delete_event(
    model: capellambse.MelodyModel,
    audit_events: list[tuple[t.Any, ...]],
    obj_id: str,
    attr: str,
    accessor_type: type,
) -> None:
    obj = model.by_uuid(obj_id)
    descriptor = getattr(type(obj), attr, None)
    assert descriptor is not None, f"{type(obj).__name__} has no {attr}"
    assert isinstance(descriptor, accessor_type), "Bad descriptor type"
    audit_events.clear()

    delattr(obj, attr)

    events = [i for i in audit_events if i[0] == "capellambse.delete"]
    assert len(events) == 1
    _, *ev = events[0]
    assert ev[0] is obj
    assert ev[1] == attr
    assert ev[2] is None



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Enumeration types used by the MelodyModel."""
import enum as _enum
import typing as t


class _StringyEnumMixin:
    """Mixin for enums that makes members compare equal to their key name."""

    name: t.Any
    _name_: t.Any

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self is other
        if isinstance(other, str):
            return self.name == other
        return NotImplemented

    def __str__(self) -> str:
        return str(self.name)

    def __hash__(self):
        return hash(self._name_)


class DiagramType(_StringyEnumMixin, _enum.Enum):
    """The types of diagrams that Capella knows about.

    Extracted from::

        $CAPELLA/eclipse/configuration/org.eclipse.osgi/635/0/.cp/description

    with::

        grep '<ownedRepresentations' *(.) \
        | grep --color=always -P '(?<=name=").*?(?=")'
    """

    UNKNOWN = "(Unknown Diagram Type)"
    # Common
    vars()["M&S"] = "Modes & States"
    MSM = "Mode State Machine"
    # Capella Architecture?
    CDI = "Contextual Component Detailed Interfaces"
    CEI = "Contextual Component External Interfaces"
    CII = "Contextual Component Internal Interfaces"
    IDB = "Interfaces Diagram Blank"
    # Requirements?
    CRI = "Contextual Capability Realization Involvement"
    CRB = "Capability Realization Blank"
    PD = "Package Dependencies"
    ID = "Interface Delegations"
    CDB = "Class Diagram Blank"
    IS = "Component Interfaces Scenario"
    ES = "Component Exchanges Scenario"
    FS = "Functional Scenario"
    SFCD = LFCD = PFCD = "Functional Chain Description"
    # State And Mode - Matrix?
    # Contextual State And Mode - Matrix?
    # Modes and States Reference Matrix?
    # Operational Analysis
    # Operational Activities - Requirements?
    OEBD = "Operational Entity Breakdown"
    OAIB = "Operational Activity Interaction Blank"
    OAB = "Operational Entity Blank"
    OABD = "Operational Activity Breakdown"
    ORB = "Operational Role Blank"
    OES = "Operational Interaction Scenario"
    OAS = "Activity Interaction Scenario"
    OPD = "Operational Process Description"
    OCB = "Operational Capabilities Blank"
    # Requirements - Operational Activities?
    COC = "Contextual Operational Capability"
    # System Analysis
    CM = "Contextual Mission"
    MB = "Missions Blank"
    CC = "Contextual Capability"
    MCB = "Missions Capabilities Blank"
    # System Functions - Requirements?
    # System Functions - Operational Activities?
    SFBD = "System Function Breakdown"
    SDFB = "System Data Flow Blank"
    SAB = "System Architecture Blank"
    CSA = "Contextual System Actors"
    # System Actor - Operational Actor?
    # Interfaces - Capabilities?
    # Interfaces - Scenarios?
    # Interfaces - Capabilities and Scenarios?
    # System/Actors - System Functions?
    # Requirements - System Functions?
    # Logical Architecture
    # Logical Functions - Requirements?
    # Logical Components - Requirements?
    # Logical Functions - System Functions?
    # Logical Components - Logical Functions?
    # Logical Architecture Requirement Refinements?
    # Logical Interface - Context Interface?
    # Logical Actor - Context Actor?
    LCBD = "Logical Component Breakdown"
    LFBD = "Logical Function Breakdown"
    LDFB = "Logical Data Flow Blank"
    LAB = "Logical Architecture Blank"
    CRR = "Capability Realization Refinement"
    # Requirements - Logical Functions?
    # Physical Architecture
    # Physical Functions - Requirements?
    # Physical Components - Requirements?
    # Physical Functions - Logical Functions?
    # Physical Components - Logical Components?
    # Physical Components - Physical Functions?
    # Physical Interface - Logical Interface?
    PFBD = "Physical Function Breakdown"
    PDFB = "Physical Data Flow Blank"
    PCBD = "Physical Component Breakdown"
    PAB = "Physical Architecture Blank"
    # Physical Actor - Logical Actor?
    # Requirements - Physical Functions?
    PPD = "Physical Path Description"
    # EPBS
    # Configuration Items - Requirements?
    # Configuration Items - Physical Artifacts?
    # EPBS Requirement Refinements?
    EAB = "EPBS Architecture Blank"
    CIBD = "Configuration Items Breakdown"


@_enum.unique
class FPortDir(_StringyEnumMixin, _enum.Flag):
    """Direction of component and function ports."""

    IN = _enum.auto()
    OUT = _enum.auto()
    INOUT = IN | OUT


@_enum.unique
class ExchangeItemType(_StringyEnumMixin, _enum.Enum):
    r"""The "TYPE" of ``ExchangeItem``\ s."""

    UNSET = _enum.auto()
    EVENT = _enum.auto()
    FLOW = _enum.auto()
    OPERATION = _enum.auto()
    SHARED_DATA = _enum.auto()


class Nature(_StringyEnumMixin, _enum.Enum):
    r"""The "NATURE" of ``PhysicalComponent``\ s."""

    NODE = _enum.auto()
    BEHAVIOR = _enum.auto()


class Kind(_StringyEnumMixin, _enum.Enum):
    r"""The "KIND" of ``PhysicalComponent``\ s."""

    UNSET = _enum.auto()
    HARDWARE = _enum.auto()
    PROCESSES = _enum.auto()
    SOFTWARE_DEPLOYMENT_UNIT = _enum.auto()
    DATA = _enum.auto()
    HARDWARE_COMPUTER = _enum.auto()
    SERVICES = _enum.auto()
    SOFTWARE_EXECUTION_UNIT = _enum.auto()
    FACILITIES = _enum.auto()
    MATERIALS = _enum.auto()
    SOFTWARE = _enum.auto()
    FIRMWARE = _enum.auto()
    PERSON = _enum.auto()
    SOFTWARE_APPLICATION = _enum.auto()


class VisibilityKind(_StringyEnumMixin, _enum.Enum):
    """Visibility kind."""

    UNSET = _enum.auto()
    PUBLIC = _enum.auto()
    PROTECTED = _enum.auto()
    PRIVATE = _enum.auto()
    PACKAGE = _enum.auto()


class CollectionKind(_StringyEnumMixin, _enum.Enum):
    ARRAY = _enum.auto()
    SEQUENCE = _enum.auto()


class UnionKind(_StringyEnumMixin, _enum.Enum):
    UNION = _enum.auto()
    VARIANT = _enum.auto()


class AggregationKind(_StringyEnumMixin, _enum.Enum):
    """Aggregation kind."""

    UNSET = _enum.auto()
    ASSOCIATION = _enum.auto()
    AGGREGATION = _enum.auto()
    COMPOSITION = _enum.auto()


class ScenarioKind(_StringyEnumMixin, _enum.Enum):
    UNSET = _enum.auto()
    DATA_FLOW = _enum.auto()
    FUNCTIONAL = _enum.auto()
    INTERACTION = _enum.auto()
    INTERFACE = _enum.auto()


class ControlNodeKind(_StringyEnumMixin, _enum.Enum):
    """ControlNode kind."""

    OR = _enum.auto()
    AND = _enum.auto()
    ITERATE = _enum.auto()



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""The 'CapellaRequirements' namespace."""
from __future__ import annotations

__all__ = [
    "CapellaIncomingRelation",
    "CapellaModule",
    "CapellaOutgoingRelation",
    "CapellaTypesFolder",
    "ElementRelationAccessor",
    "RelationsList",
    "RequirementsRelationAccessor",
]

import collections.abc as cabc
import os
import sys
import typing as t

from lxml import etree

import capellambse
import capellambse.model.common as c

from . import _requirements as rq
from . import exporter

c.XTYPE_ANCHORS[__name__] = "CapellaRequirements"


@c.xtype_handler(None)
class CapellaModule(rq.ReqIFElement):
    """A ReqIF Module that bundles multiple Requirement folders."""

    _xmltag = "ownedExtensions"

    folders = c.DirectProxyAccessor(rq.Folder, aslist=c.ElementList)
    requirements = c.DirectProxyAccessor(rq.Requirement, aslist=c.ElementList)
    type = c.AttrProxyAccessor(rq.ModuleType, "moduleType")
    attributes = rq.AttributeAccessor()

    def to_reqif(
        self,
        to: str | os.PathLike | t.IO[bytes],
        *,
        metadata: cabc.Mapping[str, t.Any] | None = None,
        pretty: bool = False,
        compress: bool | None = None,
    ) -> None:
        """Export this module as ReqIF XML.

        You can override some auto-generated metadata placed in the header
        section by passing a dictionary as ``metadata``. The following keys
        are supported. Unsupported keys are silently ignored.

        -   ``creation_time``: A ``datetime.datetime`` object that specifies
            this document's creation time. Default to the current time.
        -   ``comment``: Override the ReqIF file's comment. Defaults to a
            text derived from the model and module names.
        -   ``title``: Specify the document title. Defaults to the module's
            ``long_name``.

        Parameters
        ----------
        to
            Where to export to. Can be the name of a file, or a file-like
            object opened in binary mode.
        metadata
            A dictionary with additional metadata (see above).
        pretty
            Format the XML human-readable.
        compress
            Write compressed data (``*.reqifz``). Defaults to ``True``
            if ``target`` is a string or path-like and its name ends in
            ``.reqifz``, otherwise defaults to ``False``.
        """
        exporter.export_module(
            self, to, metadata=metadata, pretty=pretty, compress=compress
        )


@c.xtype_handler(None)
class CapellaIncomingRelation(rq.AbstractRequirementsRelation):
    """A Relation between a requirement and an object."""

    _xmltag = "ownedRelations"


@c.xtype_handler(None)
class CapellaOutgoingRelation(rq.AbstractRequirementsRelation):
    """A Relation between an object and a requirement."""

    _xmltag = "ownedExtensions"

    source = c.AttrProxyAccessor(rq.Requirement, "target")
    target = c.AttrProxyAccessor(c.GenericElement, "source")


@c.xtype_handler(None)
class CapellaTypesFolder(rq.ReqIFElement):
    _xmltag = "ownedExtensions"

    data_type_definitions = c.DirectProxyAccessor(
        c.GenericElement,
        (rq.DataTypeDefinition, rq.EnumerationDataTypeDefinition),
        aslist=c.MixedElementList,
    )
    module_types = c.DirectProxyAccessor(rq.ModuleType, aslist=c.ElementList)
    relation_types = c.DirectProxyAccessor(
        rq.RelationType, aslist=c.ElementList
    )
    requirement_types = c.DirectProxyAccessor(
        rq.RequirementType, aslist=c.ElementList
    )


class RequirementsRelationAccessor(
    c.WritableAccessor["rq.AbstractRequirementsRelation"]
):
    """Searches for requirement relations in the architecture layer."""

    __slots__ = ("aslist",)

    def __init__(self, *args, **kw) -> None:
        super().__init__(
            *args, **kw, aslist=c.ElementList, single_attr="long_name"
        )

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        rv = self._make_list(obj, self._find_relations(obj))
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv

    def __set__(self, obj, value: t.Any) -> None:
        if not isinstance(value, cabc.Iterable):
            value = (value,)

        if CapellaOutgoingRelation in [type(i) for i in value]:
            raise NotImplementedError("Cannot insert CapellaOutgoingRelations")

        if obj._constructed:
            sys.audit("capellambse.setattr", obj, self.__name__, value)

        for i in self._find_relations(obj):
            i.getparent().remove(i)
        obj._element.extend(value)

    def __delete__(self, obj) -> None:
        assert self.aslist is not None

        if getattr(obj, "_constructed", True):
            sys.audit("capellambse.delete", obj, self.__name__, None)
        for i in self._find_relations(obj):
            parent = i.getparent()
            assert parent is not None
            parent.remove(i)

    def _find_relations(self, obj) -> list[etree._Element]:
        rels = obj._model.search(
            CapellaIncomingRelation,
            rq.InternalRelation,
            CapellaOutgoingRelation,
        )
        return [i._element for i in rels if obj in (i.source, i.target)]

    def _make_list(self, parent_obj, elements):
        assert self.aslist is not None
        return self.aslist(
            parent_obj._model, elements, c.GenericElement, parent=parent_obj
        )

    def create(
        self,
        elmlist: c.ElementListCouplingMixin,
        /,
        *type_hints: str | None,
        **kw: t.Any,
    ) -> rq.InternalRelation | CapellaIncomingRelation:
        if "target" not in kw:
            raise TypeError("No `target` for new requirement relation")
        cls = self._find_relation_type(kw["target"])
        parent = elmlist._parent._element
        with elmlist._model._loader.new_uuid(parent) as uuid:
            return cls(
                elmlist._model,
                parent,
                **kw,
                source=elmlist._parent,
                uuid=uuid,
                xtype=c.build_xtype(cls),
            )

    def delete(self, elmlist, obj) -> None:
        assert self.aslist is not None
        relations = self._find_relations(elmlist._parent)
        for relation in relations:
            if relation == obj._element:
                obj.parent._model._loader.idcache_remove(relation)
                obj.parent._element.remove(relation)
                break
        else:
            raise ValueError("Cannot delete: Target object not in this list")

    def insert(
        self,
        elmlist: c.ElementListCouplingMixin,
        index: int,
        value: c.ModelObject,
    ) -> None:
        if isinstance(value, CapellaOutgoingRelation):
            parent = value.target._element
        else:
            assert isinstance(
                value, (CapellaIncomingRelation, rq.InternalRelation)
            )
            assert elmlist._parent == value.source
            parent = elmlist._parent._element

        parent.insert(index, value._element)
        elmlist._model._loader.idcache_index(value._element)

    def _find_relation_type(
        self, target: c.GenericElement
    ) -> type[rq.InternalRelation | CapellaIncomingRelation]:
        if isinstance(target, rq.Requirement):
            return rq.InternalRelation
        elif isinstance(target, rq.ReqIFElement):
            raise TypeError(
                "Cannot create relations to targets of type"
                f" {type(target).__name__}"
            )
        else:
            return rq.InternalRelation


class RelationsList(c.ElementList["rq.AbstractRequirementsRelation"]):
    def __init__(
        self,
        model: capellambse.MelodyModel,
        elements: list[etree._Element],
        elemclass: type[t.Any] | None = None,
        *,
        source: c.ModelObject,
    ) -> None:
        del elemclass
        super().__init__(model, elements, rq.AbstractRequirementsRelation)
        self._source = source

    @t.overload
    def __getitem__(self, idx: int | str) -> rq.AbstractRequirementsRelation:
        ...

    @t.overload
    def __getitem__(
        self, idx: slice
    ) -> c.ElementList[rq.AbstractRequirementsRelation]:
        ...

    def __getitem__(self, idx):
        rel = super().__getitem__(idx)
        if isinstance(rel, c.ElementList):
            return rel

        assert isinstance(rel, rq.AbstractRequirementsRelation)
        assert self._source in (rel.source, rel.target)
        if self._source == rel.source:
            return rel.target
        else:
            return rel.source

    def by_relation_type(self, reltype: str) -> RelationsList:
        matches = []
        for elm in self._elements:
            rel_elm = c.GenericElement.from_model(self._model, elm)
            assert isinstance(rel_elm, rq.AbstractRequirementsRelation)
            if rel_elm.type is not None and rel_elm.type.name == reltype:
                matches.append(elm)
        return self._newlist(matches)

    def by_relation_class(
        self, class_: t.Literal["incoming", "outgoing", "internal"]
    ) -> RelationsList:
        relation_types = {
            "incoming": CapellaIncomingRelation,
            "outgoing": CapellaOutgoingRelation,
            "internal": rq.InternalRelation,
        }
        matches: list[etree._Element] = []
        for elm in self._elements:
            rel_elm = c.GenericElement.from_model(self._model, elm)
            if isinstance(rel_elm, relation_types[class_]):
                matches.append(rel_elm._element)
        return self._newlist(matches)

    def _newlist(self, elements: list[etree._Element]) -> RelationsList:
        listtype = self._newlist_type()
        assert issubclass(listtype, RelationsList)
        return listtype(self._model, elements, source=self._source)


class ElementRelationAccessor(
    c.WritableAccessor["rq.AbstractRequirementsRelation"]
):
    """Provides access to RequirementsRelations of a GenericElement."""

    # pylint: disable=abstract-method  # Only partially implemented for now

    __slots__ = ("aslist",)

    def __init__(self) -> None:
        super().__init__(aslist=RelationsList, single_attr="long_name")

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        relations: list[etree._Element] = []
        for relation in obj._model.search(
            CapellaIncomingRelation,
            rq.InternalRelation,
            CapellaOutgoingRelation,
        ):
            if obj in (relation.source, relation.target):
                relations.append(relation._element)
        rv = self._make_list(obj, relations)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv

    def _make_list(self, parent_obj, elements):
        assert self.aslist is not None
        return self.aslist(
            parent_obj._model,
            elements,
            None,
            parent=parent_obj,
            source=parent_obj,
        )


c.set_accessor(rq.Requirement, "relations", RequirementsRelationAccessor())
c.set_accessor(rq.Requirement, "related", ElementRelationAccessor())



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Tests for aird-filters applied after rendering a diagram."""
import pytest

import capellambse
from capellambse import diagram
from capellambse import model as model_

COMP_PORT_FILTER_DIAG = "[LAB] Test Component Port Filter"
EX_ITEMS_FILTER_DIAG = "[SDFB] Test ExchangeItem Filter"
DEFAULT_ACTIVATED_FILTERS = frozenset(
    {
        "ModelExtensionFilter",
        "hide.overlappedfunctional.chains.label.filter",
        "hide.overlappedfunctional.chains.icon.filter",
    }
)
EX_ITEMS_FILTER = "show.exchange.items.filter"
NAME_AND_EX_ITEMS_FILTER = "show.functional.exchanges.exchange.items.filter"


def test_diagram_has_activated_filters(
    model_5_2: capellambse.MelodyModel,
) -> None:
    diag: model_.diagram.Diagram = model_5_2.diagrams.by_name(
        EX_ITEMS_FILTER_DIAG
    )

    assert diag.filters == DEFAULT_ACTIVATED_FILTERS


def test_add_activated_filter_to_diagram(
    model_5_2: capellambse.MelodyModel,
) -> None:
    diag: model_.diagram.Diagram = model_5_2.diagrams.by_name(
        EX_ITEMS_FILTER_DIAG
    )

    diag.filters.add(EX_ITEMS_FILTER)

    assert EX_ITEMS_FILTER in diag.filters


@pytest.mark.parametrize("filter_name", DEFAULT_ACTIVATED_FILTERS)
def test_remove_activated_filter_on_diagram(
    model_5_2: capellambse.MelodyModel, filter_name: str
) -> None:
    diag: model_.diagram.Diagram = model_5_2.diagrams.by_name(
        EX_ITEMS_FILTER_DIAG
    )

    diag.filters.remove(filter_name)

    assert diag.filters == DEFAULT_ACTIVATED_FILTERS - {filter_name}


def test_remove_activated_filter_fails_if_not_active(
    model_5_2: capellambse.MelodyModel,
) -> None:
    diag: model_.diagram.Diagram = model_5_2.diagrams.by_name(
        EX_ITEMS_FILTER_DIAG
    )

    with pytest.raises(KeyError):
        diag.filters.remove(EX_ITEMS_FILTER)


def test_component_ports_filter_is_applied(
    model: capellambse.MelodyModel,
) -> None:
    diag: model_.diagram.Diagram = model.diagrams.by_name(
        COMP_PORT_FILTER_DIAG
    )

    rendered = diag.render(None)

    assert all(
        d.hidden
        for d in rendered
        if d.port and d.styleclass and d.styleclass.startswith("CP_")
    )


def test_fex_exchangeitems_filter_is_applied(
    model_5_2: capellambse.MelodyModel,
) -> None:
    diag: model_.diagram.Diagram = model_5_2.diagrams.by_name(
        EX_ITEMS_FILTER_DIAG
    )

    diag.filters.add(EX_ITEMS_FILTER)
    rendered = diag.render(None, sorted_exchangedItems=False)
    fex_edge = next(
        ex for ex in rendered if ex.styleclass == "FunctionalExchange"
    )

    assert isinstance(fex_edge, diagram.Edge)
    assert len(fex_edge.labels) == 1
    assert fex_edge.labels[0].label == "[ExchangeItem 1, Example]"


@pytest.mark.parametrize(
    "sort,expected_label",
    [
        pytest.param(
            False,
            "Test fex [ExchangeItem 1, Example]",
            id="Keep Capella order",
        ),
        pytest.param(
            True,
            "Test fex [Example, ExchangeItem 1]",
            id="With sorted ExchangeItems",
        ),
    ],
)
def test_fex_exchangeitems_filter_with_name_is_applied(
    model_5_2: capellambse.MelodyModel, sort: bool, expected_label: str
) -> None:
    diag: model_.diagram.Diagram = model_5_2.diagrams.by_name(
        EX_ITEMS_FILTER_DIAG
    )

    diag.filters.add(NAME_AND_EX_ITEMS_FILTER)
    rendered = diag.render(None, sorted_exchangedItems=sort)
    fex_edge = rendered["_yovTvM-ZEeytxdoVf3xHjA"]

    assert isinstance(fex_edge, diagram.Edge)
    assert len(fex_edge.labels) == 1
    assert fex_edge.labels[0].label == expected_label



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Property Value Management extension for CapellaMBSE."""
from __future__ import annotations

import typing as t

from lxml import etree

import capellambse
import capellambse.model.common as c
import capellambse.pvmt.exceptions as pvexc


class PropertyValueProxy:
    # pylint: disable=line-too-long
    """Provides access to an element's property values.

    Example for accessing property values on any object that has pvmt::

        >>> model.la.all_functions[0].pvmt['domain.group.property']
        'property'
        >>> model.la.all_functions[0].pvmt['domain.group']
        <pvmt.AppliedPropertyValueGroup "domain.group"(abcdef01-2345-6789-abcd-ef0123456789)>

    .. note::
        Access is only given if the PVMT Extension is successfully
        loaded on loading the model with the :class:`MelodyModel`.
    """
    # pylint: enable=line-too-long

    _model: capellambse.MelodyModel
    _element: etree._Element

    @classmethod
    def from_model(
        cls, model: capellambse.MelodyModel, element: etree._Element
    ) -> PropertyValueProxy:
        """Create a PropertyValueProxy for an element."""
        if not hasattr(model, "_pvext") or model._pvext is None:
            raise RuntimeError("Cannot access PVMT: extension is not loaded")

        self = cls.__new__(cls)
        self._model = model
        self._element = element
        return self

    def __init__(self, **kw: t.Any) -> None:
        raise TypeError("Cannot create PropertyValueProxy this way")

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._element is other._element

    def __getitem__(self, key):
        path = key.split(".")
        if len(path) < 1 or len(path) > 3:
            raise ValueError(
                "Provide a name as `domain`, `domain.group` or"
                " `domain.group.prop`"
            )

        domain = _PVMTDomain(self._model, self._element, path[0])
        if len(path) == 1:
            return domain
        else:
            return domain[".".join(path[1:])]


class _PVMTDomain:
    def __init__(
        self,
        model: capellambse.MelodyModel,
        element: etree._Element,
        domain: str,
    ):
        self._model = model
        self._element = element
        self._domain = domain

    def __getitem__(self, key):
        path = key.split(".")
        if len(path) < 1 or len(path) > 2:
            raise ValueError("Provide a name as `group` or `group.prop`")

        try:
            pvgroup = self._model._pvext.get_element_pv(
                self._element, f"{self._domain}.{path[0]}", create=False
            )
        except pvexc.GroupNotAppliedError:
            return None

        if len(path) == 1:
            return pvgroup
        else:
            return pvgroup[path[1]]

    def __repr__(self) -> str:
        return f"<PVMTDomain {self._domain!r} on {self._model!r}>"


def init() -> None:
    c.set_accessor(
        c.GenericElement, "pvmt", c.AlternateAccessor(PropertyValueProxy)
    )



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Implementation of a ReqIF 1.1 and 1.2 exporter."""
from __future__ import annotations

import collections.abc as cabc
import contextlib
import datetime
import itertools
import math
import os
import re
import typing as t
import zipfile

from lxml import builder, etree, html

import capellambse

from . import _capellareq as cr
from . import _requirements as rq

NS = "http://www.omg.org/spec/ReqIF/20110401/reqif.xsd"
SCHEMA = "https://www.omg.org/spec/ReqIF/20110401/reqif.xsd"
XHMTL_NS = "http://www.w3.org/1999/xhtml"
XSI_NS = "http://www.w3.org/2001/XMLSchema-instance"
NSMAP = {None: NS, "xhtml": XHMTL_NS, "xsi": XSI_NS}
REQIF_UTC_DATEFORMAT = "%Y-%m-%dT%H:%M:%SZ"

STD_SPEC_OBJECT_ATTRIBUTES = {
    "ForeignID": {"type": "STRING", "attr": "identifier"},
    "ChapterName": {"type": "XHTML", "attr": "chapter_name"},
    "Name": {"type": "XHTML", "attr": "name"},
    "Text": {"type": "XHTML", "attr": "text"},
}
STD_SPECIFICATION_ATTRIBUTES = {
    "Description": {"type": "XHTML", "attr": "long_name"},
}

E = builder.ElementMaker(namespace=NS)


class _AttributeDefinition(t.NamedTuple):
    modelobj: (
        rq.AttributeDefinition | rq.AttributeDefinitionEnumeration | None
    )
    type: str

    def __hash__(self) -> int:
        if self.modelobj is None:
            return hash((None, self.type))
        return hash((self.modelobj.uuid, self.type))


def export_module(
    module: cr.CapellaModule,
    target: str | os.PathLike | t.IO[bytes],
    *,
    metadata: cabc.Mapping[str, t.Any] | None = None,
    pretty: bool = False,
    compress: bool | None = None,
) -> None:
    data = etree.Element(
        "REQ-IF",
        attrib={f"{{{XSI_NS}}}schemaLocation": f"{SCHEMA} {NS}"},
        nsmap=NSMAP,
    )
    header, timestamp = _build_header(module, metadata)
    data.append(header)
    data.append(_build_content(module, timestamp))

    if compress is None and isinstance(target, (str, os.PathLike)):
        compress = os.fspath(target).endswith(".reqifz")
    else:
        compress = False

    if not isinstance(target, (str, os.PathLike)):
        ctx: t.ContextManager[t.IO[bytes]] = contextlib.nullcontext(target)
    else:
        ctx = open(target, "wb")  # pylint: disable=consider-using-with

    with contextlib.ExitStack() as stack:
        container = stack.enter_context(ctx)
        if compress:
            archive = stack.enter_context(
                zipfile.ZipFile(container, "w", zipfile.ZIP_DEFLATED)
            )
            file = archive.open("export.reqif", "w")
        else:
            file = container
        etree.ElementTree(data).write(
            file, encoding="utf-8", pretty_print=pretty, xml_declaration=True
        )


def _build_header(
    module: cr.CapellaModule,
    metadata: cabc.Mapping[str, t.Any] | None = None,
) -> tuple[etree._Element, str]:
    if metadata is None:
        metadata = {}
    comment = metadata.get(
        "comment",
        f"Requirements module {module.name!r} from {module._model.name!r}",
    )
    title = metadata.get("title", module.long_name)
    creation_time = (
        metadata.get("creation_time", datetime.datetime.now())
        .astimezone(datetime.timezone.utc)
        .strftime(REQIF_UTC_DATEFORMAT)
    )
    header = etree.Element("REQ-IF-HEADER")
    header.set("IDENTIFIER", "_" + module._model.uuid.upper())
    for key, value in (
        ("COMMENT", comment),
        ("CREATION-TIME", creation_time),
        ("REQ-IF-TOOL-ID", f"capellambse v{capellambse.__version__}"),
        ("REQ-IF-VERSION", "1.1"),
        ("SOURCE-TOOL-ID", f"Capella {module._model.info.capella_version}"),
        ("TITLE", title),
    ):
        header.append(el := etree.Element(key))
        el.text = value
    wrapper = etree.Element("THE-HEADER")
    wrapper.append(header)
    return wrapper, creation_time


def _build_content(module: cr.CapellaModule, timestamp: str) -> etree._Element:
    content = etree.Element("REQ-IF-CONTENT")

    datatypes = etree.Element("DATATYPES")
    spec_types = etree.Element("SPEC-TYPES")
    spec_objects = etree.Element("SPEC-OBJECTS")
    spec_relations = etree.Element("SPEC-RELATIONS")
    specifications = etree.Element("SPECIFICATIONS")
    spec_relation_groups = etree.Element("SPEC-RELATION-GROUPS")

    reqtypes = _collect_objects(module)
    attr_definitions = itertools.chain.from_iterable(reqtypes.values())
    children = list(_synthesize_standard_datatypes(timestamp))
    children.extend(_build_datatypes(attr_definitions, timestamp))
    children = sorted(children, key=lambda i: i.get("IDENTIFIER"))
    datatypes.extend(children)

    children = list(
        _build_spec_object_types(module._model, reqtypes.items(), timestamp)
    )
    children = sorted(children, key=lambda i: i.get("IDENTIFIER"))
    spec_types.extend(children)

    spec_types.append(_build_specification_type(module, timestamp))

    spec_objects.extend(_build_spec_objects(module, timestamp))
    specifications.extend(_build_specifications(module, timestamp))

    content.append(datatypes)
    content.append(spec_types)
    content.append(spec_objects)
    content.append(spec_relations)
    content.append(specifications)
    content.append(spec_relation_groups)

    wrapper = etree.Element("CORE-CONTENT")
    wrapper.append(content)
    return wrapper


def _synthesize_standard_datatypes(
    timestamp: str,
) -> t.Iterable[etree._Element]:
    collected_names: list[str] = []
    for std_attributes in (
        STD_SPEC_OBJECT_ATTRIBUTES,
        STD_SPECIFICATION_ATTRIBUTES,
    ):
        for name, std_attr in std_attributes.items():
            if name in collected_names:
                continue
            collected_names.append(name)
            attributes = {
                "IDENTIFIER": f"_STD-DATATYPE-ReqIF.{name}",
                "LAST-CHANGE": timestamp,
                "LONG-NAME": f"ReqIF.{name}",
            }
            if std_attr["type"] == "STRING":
                attributes["MAX-LENGTH"] = "32000"
            elem = E(
                f"DATATYPE-DEFINITION-{std_attr['type']}",
                attributes,
            )
            yield elem


def _build_datatypes(
    attrdefs: cabc.Iterable[_AttributeDefinition], timestamp: str
) -> cabc.Iterable[etree._Element]:
    visited_types: set[str] = set()
    for attrdef, attrtype in attrdefs:
        if attrdef and attrdef.data_type:
            uuid = attrdef.data_type.uuid.upper()
        else:
            uuid = "NULL-DATATYPE"

        id = f"_{uuid}--{attrtype}"
        if id in visited_types:
            continue
        visited_types.add(id)

        elem = etree.Element(f"DATATYPE-DEFINITION-{attrtype}")
        elem.set("IDENTIFIER", id)
        elem.set("LAST-CHANGE", timestamp)
        if attrdef and attrdef.data_type:
            if attrdef.data_type.long_name:
                elem.set("LONG-NAME", attrdef.data_type.long_name)

        type_attrs = {
            "STRING": {"MAX-LENGTH": "2147483647"},
            "REAL": {"ACCURACY": "100", "MAX": "Infinity", "MIN": "-Infinity"},
            "INTEGER": {"MAX": "2147483647", "MIN": "-2147483648"},
        }
        elem.attrib.update(type_attrs.get(attrtype, {}))

        if isinstance(attrdef, rq.AttributeDefinitionEnumeration):
            values = etree.Element("SPECIFIED-VALUES")
            for i in attrdef.data_type.values:
                v = etree.Element("ENUM-VALUE")
                v.set("IDENTIFIER", "_" + i.uuid.upper())
                v.set("LAST-CHANGE", timestamp)
                if i.long_name:
                    v.set("LONG-NAME", i.long_name)
                if i.description:
                    v.set("DESC", i.description)
                values.append(v)
            elem.append(values)

        yield elem


def _build_specification_type(
    module: cr.CapellaModule, timestamp: str
) -> etree._Element:
    module_type = module.type
    identifier = (
        "NULL-SPECIFICATION-TYPE"
        if module_type is None
        else module_type.uuid.upper()
    )
    spec_attributes = E("SPEC-ATTRIBUTES")
    for name, std_attr in STD_SPECIFICATION_ATTRIBUTES.items():
        spec_attributes.append(
            E(
                f"ATTRIBUTE-DEFINITION-{std_attr['type']}",
                {
                    "IDENTIFIER": (
                        "_STD-SPECIFICATION-ATTRIBUTE-"
                        f"{identifier}"
                        f"-ReqIF.{name}"
                    ),
                    "LAST-CHANGE": timestamp,
                    "LONG-NAME": f"ReqIF.{name}",
                },
                E(
                    "TYPE",
                    E(
                        f"DATATYPE-DEFINITION-{std_attr['type']}-REF",
                        f"_STD-DATATYPE-ReqIF.{name}",
                    ),
                ),
            )
        )
    if module_type is None:
        return E(
            "SPECIFICATION-TYPE",
            {
                "IDENTIFIER": f"_{identifier}",
                "LAST-CHANGE": timestamp,
                "LONG-NAME": "Null specification type",
                "DESC": "No module type was selected in Capella.",
            },
            spec_attributes,
        )
    else:
        return E(
            "SPECIFICATION-TYPE",
            {
                "IDENTIFIER": f"_{identifier}",
                "LAST-CHANGE": timestamp,
                "LONG-NAME": module_type.long_name,
            },
            spec_attributes,
        )


def _build_spec_object_types(
    model: capellambse.MelodyModel,
    reqtypes: cabc.Iterable[tuple[str | None, set[_AttributeDefinition]]],
    timestamp: str,
) -> cabc.Iterable[etree._Element]:
    for reqtype, attr_defs in reqtypes:
        modelobj: rq.RequirementType | None
        if reqtype:
            modelobj = model.by_uuid(reqtype)  # type: ignore[assignment]
            assert isinstance(modelobj, rq.RequirementType)
            reqtype = reqtype.upper()
        else:
            modelobj = None
            reqtype = "NULL-SPEC-OBJECT-TYPE"

        elem = etree.Element("SPEC-OBJECT-TYPE")
        elem.set("IDENTIFIER", "_" + reqtype)
        elem.set("LAST-CHANGE", timestamp)
        if not modelobj:
            elem.set("LONG-NAME", "Null spec object type")
            elem.set("DESC", "No requirement type was selected in Capella.")
        else:
            _add_common_attributes(modelobj, elem)

        attributes_wrap = etree.Element("SPEC-ATTRIBUTES")
        for name, std_attr in STD_SPEC_OBJECT_ATTRIBUTES.items():
            attributes = {
                "IDENTIFIER": f"_STD-ATTRIBUTE-{reqtype}-ReqIF.{name}",
                "LAST-CHANGE": timestamp,
                "LONG-NAME": f"ReqIF.{name}",
            }
            elem_attrdef = E(
                f"ATTRIBUTE-DEFINITION-{std_attr['type']}",
                attributes,
                E(
                    "TYPE",
                    E(
                        f"DATATYPE-DEFINITION-{std_attr['type']}-REF",
                        f"_STD-DATATYPE-ReqIF.{name}",
                    ),
                ),
            )
            attributes_wrap.append(elem_attrdef)
        for attr_def in attr_defs:
            attr_elem = etree.Element(f"ATTRIBUTE-DEFINITION-{attr_def.type}")
            if attr_def.modelobj:
                attid = attr_def.modelobj.uuid.upper()
                dt = attr_def.modelobj.data_type
                dtid = dt.uuid.upper() if dt else "NULL-DATATYPE"
            else:
                attid = "NULL-ATTRIBUTE-DEFINITION"
                dtid = "NULL-DATATYPE"
            attr_elem.set("IDENTIFIER", f"_{attid}--{attr_def.type}")
            attr_elem.set("LAST-CHANGE", timestamp)
            _add_common_attributes(attr_def.modelobj, attr_elem)
            if attr_def.type == "ENUMERATION":
                assert attr_def.modelobj is not None
                attr_elem.set(
                    "MULTI-VALUED",
                    ("false", "true")[attr_def.modelobj.multi_valued],
                )

            dtref = etree.Element(f"DATATYPE-DEFINITION-{attr_def.type}-REF")
            dtref.text = f"_{dtid}--{attr_def.type}"
            dtwrap = etree.Element("TYPE")
            dtwrap.append(dtref)
            attr_elem.append(dtwrap)
            attributes_wrap.append(attr_elem)
        if len(attributes_wrap):
            elem.append(attributes_wrap)

        yield elem


def _build_spec_objects(
    parent: cr.CapellaModule | rq.Folder,
    timestamp: str,
) -> cabc.Iterable[etree._Element]:
    for req in parent.requirements:
        yield _build_spec_object(req, timestamp)

    for folder in parent.folders:
        yield from _build_spec_objects(folder, timestamp)


def _build_spec_object(req: rq.Requirement, timestamp: str) -> etree._Element:
    obj = etree.Element("SPEC-OBJECT")
    obj.set("IDENTIFIER", "_" + req.uuid.upper())
    obj.set("LAST-CHANGE", timestamp)
    if req.long_name:
        obj.set("LONG-NAME", req.long_name)

    values_elem = E("VALUES")
    values_elem.extend(_build_standard_attribute_values(req))
    if req.attributes:
        values_elem.extend(_build_attribute_values(req))
    obj.append(values_elem)

    obj.append(type := etree.Element("TYPE"))
    type.append(ref := etree.Element("SPEC-OBJECT-TYPE-REF"))
    if req.type:
        ref.text = "_" + req.type.uuid.upper()
    else:
        ref.text = "_NULL-SPEC-OBJECT-TYPE"

    return obj


def _build_standard_attribute_values(
    req: rq.Requirement,
) -> t.Iterable[etree._Element]:
    if req.type:
        reqtype_ref = req.type.uuid.upper()
    else:
        reqtype_ref = "NULL-SPEC-OBJECT-TYPE"
    for name, std_attr in STD_SPEC_OBJECT_ATTRIBUTES.items():
        type_ = std_attr["type"]
        attrdef_ref = f"_STD-ATTRIBUTE-{reqtype_ref}-ReqIF.{name}"
        value_elem = None
        capella_val = getattr(req, std_attr["attr"])
        if capella_val is None:
            capella_val = ""
        if type_ == "STRING":
            value_elem = E(
                "ATTRIBUTE-VALUE-STRING",
                {"THE-VALUE": capella_val},
                E(
                    "DEFINITION",
                    E("ATTRIBUTE-DEFINITION-STRING-REF", attrdef_ref),
                ),
            )
        elif type_ == "XHTML":
            if capella_val:
                html_val = capella_val
            else:
                html_val = "<div></div>"
            xml_elem = html.fromstring(html_val)
            html.html_to_xhtml(xml_elem)
            value_elem = E(
                "ATTRIBUTE-VALUE-XHTML",
                E(
                    "DEFINITION",
                    E("ATTRIBUTE-DEFINITION-XHTML-REF", attrdef_ref),
                ),
                E("THE-VALUE", xml_elem),
            )
        if value_elem is not None:
            yield value_elem


def _build_attribute_values(
    req: rq.Requirement,
) -> t.Iterable[etree._Element]:
    for attr in req.attributes:
        if isinstance(attr, rq.EnumerationValueAttribute):
            yield _build_attribute_value_enum(attr)
        else:
            yield _build_attribute_value_simple(attr)


def _build_attribute_value_simple(
    attr: rq.AbstractRequirementsAttribute,
) -> etree._Element:
    attrtype = _attrtype2reqif(attr)
    obj = etree.Element(f"ATTRIBUTE-VALUE-{attrtype}")
    value: t.Any
    if attrtype == "BOOLEAN":
        obj.set("THE-VALUE", "true" if attr.value else "false")
    elif attrtype == "DATE":
        value = attr.value
        if value is None:
            obj.set("THE-VALUE", "1990-01-01T00:00:00Z")
        elif isinstance(value, datetime.datetime):
            utcval = value.astimezone(datetime.timezone.utc)
            obj.set("THE-VALUE", utcval.strftime(REQIF_UTC_DATEFORMAT))
        else:
            raise TypeError(f"Expected datetime, got {type(value).__name__}")
    elif attrtype == "INTEGER":
        obj.set("THE-VALUE", str(int(attr.value)))
    elif attrtype == "REAL":
        value = attr.value
        if value == math.inf:
            value = "Infinity"
        elif value == -math.inf:
            value = "-Infinity"
        else:
            value = str(value)
        obj.set("THE-VALUE", value)
    elif attrtype == "STRING":
        obj.set("THE-VALUE", attr.value or "")
    else:
        raise ValueError(f"Unknown attribute type {attrtype}")
    obj.append(_ref_attribute_definition(attrtype, attr.definition))
    return obj


def _build_attribute_value_enum(
    attribute: rq.AbstractRequirementsAttribute,
) -> etree._Element:
    obj = etree.Element("ATTRIBUTE-VALUE-ENUMERATION")

    obj.append(_ref_attribute_definition("ENUMERATION", attribute.definition))

    obj.append(values := etree.Element("VALUES"))
    for i in attribute.values:
        values.append(ref := etree.Element("ENUM-VALUE-REF"))
        ref.text = "_" + i.uuid.upper()
    return obj


def _build_specifications(
    module: cr.CapellaModule, timestamp: str
) -> cabc.Iterable[etree._Element]:
    def create_hierarchy_object(req: rq.Requirement) -> etree._Element:
        id = "_" + req.uuid.upper()
        id_hier = id + "--HIER"
        return E(
            "SPEC-HIERARCHY",
            {"IDENTIFIER": id_hier, "LAST-CHANGE": timestamp},
            E("OBJECT", E("SPEC-OBJECT-REF", id)),
        )

    def create_hierarchy_folder(
        folder: cr.CapellaModule | rq.Folder,
    ) -> cabc.Iterable[etree._Element]:
        for req in folder.requirements:
            yield create_hierarchy_object(req)
        for sub in folder.folders:
            yield from create_hierarchy_folder(sub)

    module_type_ref = (
        "NULL-SPECIFICATION-TYPE"
        if (module_type := module.type) is None
        else module_type.uuid.upper()
    )

    spec_values = E("VALUES")

    for name, std_attr in STD_SPECIFICATION_ATTRIBUTES.items():
        attr_def_ref = (
            "_STD-SPECIFICATION-ATTRIBUTE-"
            f"{module_type_ref}"
            f"-ReqIF.{name}"
        )
        if (type_ := std_attr["type"]) == "XHTML":
            xml_elem = html.fromstring(
                f"<div>{getattr(module, std_attr['attr'])}</div>"
            )
            html.html_to_xhtml(xml_elem)
            spec_values.append(
                E(
                    "ATTRIBUTE-VALUE-XHTML",
                    E.DEFINITION(
                        E("ATTRIBUTE-DEFINITION-XHTML-REF", attr_def_ref)
                    ),
                    E("THE-VALUE", xml_elem),
                )
            )
        else:
            spec_values.append(
                E(
                    f"ATTRIBUTE-VALUE-{type_}",
                    {"THE-VALUE": str(getattr(module, std_attr["attr"]))},
                    E.DEFINITION(
                        E(f"ATTRIBUTE-DEFINITION-{type_}-REF", attr_def_ref)
                    ),
                )
            )

    spec = E.SPECIFICATION(
        {
            "IDENTIFIER": "_" + module.uuid.upper(),
            "LAST-CHANGE": timestamp,
        },
        E.TYPE(E("SPECIFICATION-TYPE-REF", f"_{module_type_ref}")),
        spec_values,
        children := E.CHILDREN(),
    )
    if module.long_name:
        spec.set("LONG-NAME", module.long_name)
    if module.description:
        spec.set("DESC", module.description)

    children.extend(create_hierarchy_folder(module))
    return (spec,)


def _ref_attribute_definition(
    type: str, adef: rq.AttributeDefinition | None
) -> etree._Element:
    if adef is not None:
        reftext = f"_{adef.uuid.upper()}--{type}"
    else:
        reftext = f"NULLTYPE--{type}"
    definition = etree.Element("DEFINITION")
    ref = etree.Element(f"ATTRIBUTE-DEFINITION-{type}-REF")
    ref.text = reftext
    definition.append(ref)
    return definition


def _add_common_attributes(obj: t.Any | None, element: etree._Element) -> None:
    if obj is None:
        return

    if value := obj.long_name:
        element.set("LONG-NAME", value)
    if value := obj.description:
        element.set("DESC", value)


def _collect_objects(
    module: cr.CapellaModule,
) -> dict[str | None, set[_AttributeDefinition]]:
    requirements: set[str] = set()
    req_types: dict[str | None, set[_AttributeDefinition]] = {}

    def collect_requirement(i: rq.Requirement) -> None:
        if i.uuid in requirements:
            return
        requirements.add(i.uuid)
        attr_definitions = req_types.setdefault(i.type and i.type.uuid, set())
        for attr in i.attributes:
            adef = _AttributeDefinition(attr.definition, _attrtype2reqif(attr))
            attr_definitions.add(adef)

    def collect_folder(
        i: rq.Folder | cr.CapellaModule,
    ) -> None:
        for req in i.requirements:
            collect_requirement(req)
        for fld in i.folders:
            collect_folder(fld)

    collect_folder(module)
    return req_types


def _attrtype2reqif(
    attrtype: rq.AbstractRequirementsAttribute,
) -> str:
    m = re.fullmatch("([A-Z][A-Za-z]*)ValueAttribute", type(attrtype).__name__)
    assert m and m.group(1)
    return m.group(1).upper()



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Collection of tools for drawing model complexity assessment badge."""
from __future__ import annotations

import collections.abc as cabc

COLORS = ("#ffdd87", "#91cc84", "#a5c2e6", "#f89f9f")
LEGEND = (
    (2, "Operational Analysis"),
    (36, "System Analysis"),
    (66, "Logical Architecture"),
    (99, "Physical Architecture"),
)
"""The x-offset and label of each legend entry."""


def draw_rect(
    x: float,
    y: float,
    width: float,
    height: float,
    fill: str = "#FFF",
    stroke: str = "#333",
    stroke_width: float = 0,
) -> str:
    """Construct a rectangle (SVG string)."""
    return (
        f'<rect fill="{fill}" stroke="{stroke}" stroke-width="{stroke_width}"'
        f' x="{x}" y="{y}" width="{width}" height="{height}" />\n'
    )


def draw_group(
    contents: list[str],
    fill: str | None = None,
    font_size: float | None = None,
    font_family: str | None = None,
) -> str:
    """Construct a group (SVG string)."""
    return (
        "<g"
        + (f' font-family="{font_family}"' if font_family else "")
        + (f' font-size="{font_size:.2}"' if font_size else "")
        + (f' fill="{fill}"' if fill else "")
        + f'>\n{"".join(contents)}</g>\n'
    )


def draw_text(
    x: float, y: float, text: str, font_size: float | None = None
) -> str:
    """Construct a text element (SVG string)."""
    return (
        "<text "
        + (f'font-size="{font_size:.2}"' if font_size else "")
        + f'x="{x}" y="{y}">{text}</text>\n'
    )


def draw_legend(x: float, y: float):
    """Construct plot legend (SVG string)."""
    return draw_group(
        [
            draw_rect(x + x_offset, y, 5, 3, fill=COLORS[idx])
            + draw_text(x + x_offset + 6, y + 2.6, text)
            for idx, (x_offset, text) in enumerate(LEGEND)
        ],
        fill="#555",
        font_size=2.8,
    )


def draw_bar(
    data: cabc.Sequence[float],
    max_width: float,
    x: float,
    y: float,
    height: float = 8,
    show_label_threshold: float = 0.1,
) -> str:
    """Construct a 4-segment bar plot (SVG string).

    Segment spacing is defined by {data}; Sum of {data} must match 1 for
    the thing to work the right way. Segments are labeled with "%" if
    the width is above {show_label_threshold}.
    """
    widths = [max_width * x for x in data]
    offsets = [0.0]
    for i, width in enumerate(widths[:-1]):
        offsets.append(offsets[i] + width)
    label = [
        f"{x*100.0:.0f}%" if x >= show_label_threshold else "" for x in data
    ]
    return draw_group(
        [
            draw_rect(
                x=x + x_offset,
                y=y,
                width=widths[idx],
                height=height,
                fill=COLORS[idx],
            )
            + (
                draw_text(
                    x=x + x_offset + 1,
                    y=y + (height / 2.0) + 1,
                    text=label[idx],
                )
                if len(label[idx]) > 0
                else ""
            )
            for idx, x_offset in enumerate(offsets)
        ],
        font_size=3.2,
    )


def draw_diagrams_icon(x: float | int, y: float | int):
    """Create simple diagram icon (SVG string)."""
    return (
        '<g stroke="#555" stroke-width=".2">'
        f'<path fill="#FFF" d="M{x + 0.5} {y + 1.8} h3.6 v3.1 H{x + 0.5} z"/>'
        f'<path fill="#a5c2e6" d="M{x + 1} {y + 2.4} h1v.7 H{x + 1}z '
        'm.8 1.2h1v.7h-1z m.9-1.1h.8v.6h-.8z"/><path fill="none" '
        f'd="M{x + 1}.9 {y + 2.8} h.8 m-1.3.3.4.8 m1 .1.4-.9"/></g>'
    )


def draw_objects_icon(x: float | int, y: float | int):
    """Create simple objects icon (SVG string)."""
    return (
        '<g stroke="#555" stroke-width=".3">'
        f'<path fill="#ffdd87" d="M{x + 1} {y + 2.1} h1.9 V{y + 4} h-2 z"/>'
        f'<path fill="#a5c2e6" d="M{x + 1.8} {y + 3} h2 v1.8 h-2 z"/></g>'
    )


def draw_labeled_bar(
    x: float,
    y: float,
    label: str,
    segments: list[int],
    draw_icon: cabc.Callable[[float | int, float | int], str],
) -> str:
    """Construct a bar with label and icon (SVG string)."""
    total = sum(segments)
    normalized_segments = [x / total for x in segments]
    return (
        draw_icon(x, y)
        + draw_text(x + 5, y + 3, f"{total}")
        + draw_text(x + 5, y + 6, label)
        + draw_bar(normalized_segments, 104, x + 22.2, y)
    )


def draw_summary_badge(
    objects: list[int], diagrams: list[int], scale: float = 4.0
) -> str:
    """Construct summary badge (SVG string)."""
    text = draw_group(
        [
            draw_rect(1, 0, 132, 30, stroke_width=0.2),
            draw_rect(23.8, 1.4, 108, 22, stroke_width=0.2),
            draw_labeled_bar(2, 3.4, "objects", objects, draw_objects_icon),
            draw_labeled_bar(
                2, 13.4, "diagrams", diagrams, draw_diagrams_icon
            ),
            draw_legend(1, 25),
        ],
        font_family="sans-serif",
        font_size=3.2,
    )
    return (
        f'<SVG xmlns="http://www.w3.org/2000/svg"'
        f' width="{scale*134}" height="{scale*30}" viewBox="0 0 134 30">'
        f"{text}</SVG>"
    )



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Tools for the Operational Analysis layer.

.. diagram:: [CDB] OA ORM
"""
from __future__ import annotations

import operator

from .. import common as c
from .. import crosslayer, diagram
from ..crosslayer import (
    capellacommon,
    capellacore,
    cs,
    fa,
    information,
    interaction,
)

XT_ARCH = "org.polarsys.capella.core.data.oa:OperationalAnalysis"


@c.xtype_handler(XT_ARCH)
class OperationalActivity(fa.AbstractFunction):
    """An operational activity."""

    _xmltag = "ownedOperationalActivities"

    owning_entity = c.CustomAccessor(
        c.GenericElement,
        operator.attrgetter("_model.oa.all_entities"),
        matchtransform=operator.attrgetter("activities"),
    )
    exchanges = c.DirectProxyAccessor(
        fa.FunctionalExchange, aslist=c.ElementList
    )

    @property
    def inputs(self) -> c.ElementList[fa.FunctionalExchange]:
        return self._model.oa.all_activity_exchanges.by_target(self)

    @property
    def outputs(self) -> c.ElementList[fa.FunctionalExchange]:
        return self._model.oa.all_activity_exchanges.by_source(self)

    @property
    def related_exchanges(self) -> c.ElementList[fa.FunctionalExchange]:
        seen: set[str] = set()
        exchanges = []
        for fex in self.inputs + self.outputs:
            if fex.uuid not in seen:
                exchanges.append(fex._element)
                seen.add(fex.uuid)
        return self.inputs._newlist(exchanges)


@c.xtype_handler(XT_ARCH)
class OperationalProcess(fa.FunctionalChain):
    """An operational process."""


@c.xtype_handler(XT_ARCH)
class EntityOperationalCapabilityInvolvement(interaction.AbstractInvolvement):
    """An EntityOperationalCapabilityInvolvement."""


@c.xtype_handler(XT_ARCH)
class OperationalCapability(c.GenericElement):
    """A capability in the OperationalAnalysis layer."""

    _xmltag = "ownedOperationalCapabilities"

    extends = c.DirectProxyAccessor(
        interaction.AbstractCapabilityExtend, aslist=c.ElementList
    )
    extended_by = c.ReferenceSearchingAccessor(
        interaction.AbstractCapabilityExtend, "target", aslist=c.ElementList
    )
    includes = c.DirectProxyAccessor(
        interaction.AbstractCapabilityInclude, aslist=c.ElementList
    )
    included_by = c.ReferenceSearchingAccessor(
        interaction.AbstractCapabilityInclude, "target", aslist=c.ElementList
    )
    generalizes = c.DirectProxyAccessor(
        interaction.AbstractCapabilityGeneralization, aslist=c.ElementList
    )
    generalized_by = c.DirectProxyAccessor(
        interaction.AbstractCapabilityGeneralization,
        "target",
        aslist=c.ElementList,
    )
    involved_activities = c.LinkAccessor[OperationalActivity](
        "ownedAbstractFunctionAbstractCapabilityInvolvements",
        interaction.AbstractFunctionAbstractCapabilityInvolvement,
        aslist=c.ElementList,
        attr="involved",
    )
    involved_entities = c.LinkAccessor[c.GenericElement](
        "ownedEntityOperationalCapabilityInvolvements",
        EntityOperationalCapabilityInvolvement,
        aslist=c.MixedElementList,
        attr="involved",
    )
    entity_involvements = c.DirectProxyAccessor(
        EntityOperationalCapabilityInvolvement, aslist=c.ElementList
    )
    involved_processes = c.LinkAccessor[OperationalProcess](
        "ownedFunctionalChainAbstractCapabilityInvolvements",
        interaction.XT_CAP2PROC,
        aslist=c.ElementList,
        attr="involved",
    )
    owned_processes = c.DirectProxyAccessor(
        OperationalProcess, aslist=c.ElementList
    )

    postcondition = c.AttrProxyAccessor(
        capellacore.Constraint, "postCondition"
    )
    precondition = c.AttrProxyAccessor(capellacore.Constraint, "preCondition")
    scenarios = c.DirectProxyAccessor(
        interaction.Scenario, aslist=c.ElementList
    )
    states = c.AttrProxyAccessor(
        capellacommon.State, "availableInStates", aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class OperationalCapabilityPkg(c.GenericElement):
    """A package that holds operational capabilities."""

    _xmltag = "ownedAbstractCapabilityPkg"

    capabilities = c.DirectProxyAccessor(
        OperationalCapability, aslist=c.ElementList
    )

    packages: c.Accessor


class AbstractEntity(cs.Component):
    """Common code for Entities."""

    activities = c.LinkAccessor[OperationalActivity](
        "ownedFunctionalAllocation",
        fa.XT_FCALLOC,
        aslist=c.ElementList,
        attr="targetElement",
    )
    capabilities = c.CustomAccessor(
        OperationalCapability,
        operator.attrgetter("_model.oa.all_capabilities"),
        matchtransform=operator.attrgetter("involved_entities"),
        aslist=c.ElementList,
    )


@c.xtype_handler(XT_ARCH)
class Entity(AbstractEntity):
    """An Entity in the OperationalAnalysis layer."""

    _xmltag = "ownedEntities"

    entities: c.Accessor

    @property
    def inputs(self) -> c.ElementList[CommunicationMean]:
        return self._model.oa.all_entity_exchanges.by_target(self)

    @property
    def outputs(self) -> c.ElementList[CommunicationMean]:
        return self._model.oa.all_entity_exchanges.by_source(self)


@c.xtype_handler(XT_ARCH)
class OperationalActivityPkg(c.GenericElement):
    """A package that holds operational entities."""

    _xmltag = "ownedFunctionPkg"

    activities = c.DirectProxyAccessor(
        OperationalActivity, aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class CommunicationMean(fa.AbstractExchange):
    """An operational entity exchange."""

    _xmltag = "ownedComponentExchanges"

    allocated_interactions = c.LinkAccessor[fa.FunctionalExchange](
        None,  # FIXME fill in tag
        fa.XT_COMP_EX_FNC_EX_ALLOC,
        aslist=c.ElementList,
        attr="targetElement",
    )
    allocated_exchange_items = c.AttrProxyAccessor(
        information.ExchangeItem,
        "convoyedInformations",
        aslist=c.ElementList,
    )

    exchange_items = fa.ComponentExchange.exchange_items


@c.xtype_handler(XT_ARCH)
class EntityPkg(c.GenericElement):
    """A package that holds operational entities."""

    _xmltag = "ownedEntityPkg"

    entities = c.DirectProxyAccessor(Entity, aslist=c.ElementList)
    state_machines = c.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )

    packages: c.Accessor
    exchanges = c.DirectProxyAccessor(CommunicationMean, aslist=c.ElementList)


@c.xtype_handler(None)
class OperationalAnalysis(crosslayer.BaseArchitectureLayer):
    """Provides access to the OperationalAnalysis layer of the model."""

    root_entity = c.DirectProxyAccessor(Entity, rootelem=EntityPkg)
    root_activity = c.DirectProxyAccessor(
        OperationalActivity, rootelem=OperationalActivityPkg
    )

    activity_package = c.DirectProxyAccessor(OperationalActivityPkg)
    capability_package = c.DirectProxyAccessor(OperationalCapabilityPkg)
    entity_package = c.DirectProxyAccessor(EntityPkg)

    all_activities = c.DeepProxyAccessor(
        OperationalActivity,
        aslist=c.ElementList,
    )
    all_processes = c.DeepProxyAccessor(
        OperationalProcess,
        aslist=c.ElementList,
    )
    all_capabilities = c.DeepProxyAccessor(
        OperationalCapability,
        aslist=c.ElementList,
    )
    all_actors = property(
        lambda self: self._model.search(Entity).by_is_actor(True)
    )
    all_entities = c.DeepProxyAccessor(
        Entity,
        aslist=c.ElementList,
    )

    all_activity_exchanges = c.DeepProxyAccessor(
        fa.FunctionalExchange,
        aslist=c.ElementList,
        rootelem=[OperationalActivityPkg, OperationalActivity],
    )
    all_entity_exchanges = c.DeepProxyAccessor(
        CommunicationMean,
        aslist=c.ElementList,
    )

    diagrams = diagram.DiagramAccessor(
        "Operational Analysis", cacheattr="_MelodyModel__diagram_cache"
    )


c.set_accessor(
    OperationalActivity,
    "packages",
    c.DirectProxyAccessor(OperationalActivityPkg, aslist=c.ElementList),
)
c.set_accessor(
    Entity,
    "exchanges",
    c.DirectProxyAccessor(CommunicationMean, aslist=c.ElementList),
)
c.set_accessor(
    Entity,
    "related_exchanges",
    c.ReferenceSearchingAccessor(
        CommunicationMean, "source", "target", aslist=c.ElementList
    ),
)
c.set_self_references(
    (OperationalActivity, "activities"),
    (OperationalActivityPkg, "packages"),
    (OperationalCapabilityPkg, "packages"),
    (Entity, "entities"),
    (EntityPkg, "packages"),
)



#!/usr/bin/env -S python -Xdev
# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""A simple Python REPL with a Capella model.

You can use this script to quickly launch an interactive Python
interpreter and load a model.

If the ``readline`` module is available, the interpreter spawned from
this script uses a separate readline history. It is located in
``$XDG_STATE_HOME/capellambse`` on proper operating systems and in the
``capellambse`` cache directory on others.

Normally this script is run with something like ``python -Xdev -m
capellambse.repl test-5.0``. However, as that can become quite unwieldy,
it is also possible to run it as ``./capellambse/repl.py 5.0`` from the
source tree – on Unix-like operating systems, this will automatically
enable ``-Xdev`` on the Python interpreter. Requirement is a
sufficiently recent version of ``env``, which by now even Debian should
have.

In order to add custom models for the :ref:`model <repl.py-model>`
argument, add a JSON file to the ``capellambse/known_models`` directory.
This file defines the instantiation parameters for the
:class:`capellambse.model.MelodyModel`:

.. literalinclude:: ../../../capellambse/known_models/test-lib.json
   :language: json
   :lineno-start: 1
   :linenos:
"""
from __future__ import annotations

import argparse
import code
import collections.abc as cabc
import contextlib
import importlib
import importlib.resources as imr
import json
import logging
import os
import os.path
import pathlib
import sys
import textwrap
import typing as t

from lxml import etree

import capellambse
from capellambse.loader import exs

try:
    import readline

    try:
        import rlcompleter  # pylint: disable=unused-import
    except ImportError:
        pass

    class _ReadlineHistory:
        def __init__(self, histfile):
            self.histfile = histfile

        def __enter__(self) -> None:
            try:
                readline.read_history_file(self.histfile)
            except FileNotFoundError:
                pass
            readline.parse_and_bind("tab: complete")

        def __exit__(self, *_) -> None:
            try:
                self.histfile.parent.mkdir(parents=True, exist_ok=True)
                self.histfile.open("wb").close()
            except OSError:
                pass
            readline.append_history_file(100_000, self.histfile)

except ImportError:

    class _ReadlineHistory:  # type: ignore[no-redef]
        def __init__(self, *__, **_):
            pass

        def __enter__(self):
            pass

        def __exit__(self, *_):
            pass


_ModelInfo = t.Dict[str, t.Union[str, t.Dict[str, str]]]

basedir = pathlib.Path(__file__).parent.resolve()
logger = logging.getLogger("capellambse.repl")


def _load_model_info(datapath: str | importlib.abc.Traversable) -> _ModelInfo:
    if isinstance(datapath, str):
        try:
            return json.loads(datapath)
        except json.JSONDecodeError:
            pass
        datapath = imr.files(capellambse) / "known_models" / f"{datapath}.json"
    with datapath.open("r", encoding="utf-8") as file:
        return json.load(file)


def _parse_args(args: list[str] | None = None) -> dict[str, t.Any]:
    known_models = list(
        i.name[: -len(".json")]
        for i in capellambse.enumerate_known_models()
        if i.name.endswith(".json")
    )
    parser = argparse.ArgumentParser("capellambse/repl.py")
    parser.add_argument(
        "model",
        default="test-5.0",
        nargs=1,
        help=(
            "A model name from known_models, an AIRD file, or the path to or"
            " contents of a JSON file describing the model. The following"
            " models are known: " + ", ".join(f"``{i}``" for i in known_models)
        ),
    )
    parser.add_argument(
        "--disable-diagram-cache",
        action="store_true",
        help="Disable the diagram cache, if one was defined for the model",
    )
    parser.add_argument(
        "--dump",
        action="store_true",
        help="Dump model info as JSON to stdout and exit",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"capellambse {capellambse.__version__}",
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--hold",
        action="store_true",
        help="Inhibit automatic model updates (update_cache=False)",
    )
    group.add_argument(
        "--wipe",
        action="store_true",
        help="Wipe the cache (disable_cache=True)",
    )

    pargs = parser.parse_args(args or sys.argv[1:])
    assert len(pargs.model) == 1
    (model,) = pargs.model
    modelinfo: dict[str, t.Any]
    if model.endswith(".json"):
        modelinfo = _load_model_info(pathlib.Path(model))
    elif model.endswith(".aird"):
        modelinfo = {"path": model}
    else:
        modelinfo = _load_model_info(model)

    if pargs.hold:
        modelinfo["update_cache"] = False
    if pargs.wipe:
        modelinfo["disable_cache"] = True

    if pargs.disable_diagram_cache and "diagram_cache" in modelinfo:
        del modelinfo["diagram_cache"]

    if pargs.dump:
        print(json.dumps(modelinfo, indent=2))
        raise SystemExit(0)
    return modelinfo


@contextlib.contextmanager
def logtee(
    filename: os.PathLike[str] | str,
    /,
    *,
    append: bool = False,
    catch: bool = False,
    tee: bool = True,
    level: int | str | None = None,
    logger: str = "",  # pylint: disable=redefined-outer-name
) -> cabc.Generator[None, None, None]:
    """Temporarily write log messages into ``file``.

    Parameters
    ----------
    file
        The file path to write log messages to.
    append
        Append to the file instead of overwriting it.
    catch
        If ``True``, catches and suppresses exceptions that reach the
        context manager. Only ``Exception`` and subtypes will be
        suppressed. Regardless of this parameter, exceptions (of any
        type) will always be logged (subject to the log level).
    level
        Change the log level during redirection.
    logger
        Attach to this sub-logger instead of the root logger.
    tee
        Keep the original log destination(s) intact as well. If
        ``False``, also inhibits propagation of log messages to parent
        loggers.
    """
    loggerobj = logging.getLogger(logger)
    handler = logging.FileHandler(filename, mode="wa"[bool(append)])

    orig_handlers = loggerobj.handlers
    orig_propagate = loggerobj.propagate
    orig_level = loggerobj.level

    loggerobj.handlers = orig_handlers * bool(tee) + [handler]
    loggerobj.propagate = bool(tee)
    if level is not None:
        loggerobj.setLevel(level)

    try:
        yield
    except BaseException as err:
        loggerobj.handlers = [handler]
        loggerobj.propagate = False
        loggerobj.exception("Exception reached top of context")
        if not isinstance(err, Exception) or not catch:
            raise
    finally:
        loggerobj.handlers = orig_handlers
        loggerobj.propagate = orig_propagate
        loggerobj.level = orig_level
        handler.close()


@contextlib.contextmanager
def suppress(
    *exc_types: type[BaseException], log: bool = True
) -> cabc.Iterator[None]:
    """Prevent the specified types of exceptions from propagating.

    Parameters
    ----------
    exc_types
        The exception types to suppress. Subclasses of these exceptions
        will also be suppressed.
    log
        Print a short warning about the exception to stderr.
    """
    try:
        yield
    except exc_types:
        if log:
            logging.info("Exception suppressed", exc_info=True)


def showxml(obj: capellambse.ModelObject | etree._Element) -> None:
    if isinstance(obj, etree._Element):
        elm = obj
    else:
        elm = obj._element
    print(exs.to_string(elm), end="")


def main() -> None:
    """Launch a simple Python REPL with a Capella model."""
    os.chdir(pathlib.Path(capellambse.__file__).parents[1])

    modelinfo = _parse_args()
    logger.debug("Loading model: %r", modelinfo)
    model = capellambse.MelodyModel(**modelinfo)

    interactive_locals = {
        "__doc__": None,
        "__name__": "__console__",
        "etree": etree,
        "im": importlib,
        "imm": importlib.import_module("importlib.metadata"),
        "imr": importlib.import_module("importlib.resources"),
        "logtee": logtee,
        "model": model,
        "pprint": importlib.import_module("pprint").pprint,
        "showxml": showxml,
        "suppress": suppress,
    }

    for m in ("capellambse", "inspect", "logging", "os", "pathlib"):
        interactive_locals[m] = importlib.import_module(m)

    banner = textwrap.dedent(
        f"""\

        {' Model exploration ':=^80}
        `model` is {model.info.title!r} from {modelinfo['path']}
        """
    )

    history_file = capellambse.dirs.user_state_path / "model_exploration.hist"
    with _ReadlineHistory(history_file), suppress(BrokenPipeError):
        code.interact(banner=banner, local=interactive_locals, exitmsg="")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Tests for creating and deleting model elements."""
# pylint: disable=missing-function-docstring, redefined-outer-name
import pathlib

import pytest

import capellambse

TEST_ROOT = pathlib.Path(__file__).parent / "data" / "writemodel"
TEST_MODEL = "WriteTestModel.aird"

XPATH_UUID = "//*[@id={!r}]"


@pytest.fixture
def model():
    return capellambse.MelodyModel(TEST_ROOT / TEST_MODEL)


def test_created_elements_can_be_accessed_in_model(
    model: capellambse.MelodyModel,
):
    newobj = model.la.root_component.components.create(name="TestComponent")

    assert newobj is not None
    assert isinstance(newobj, capellambse.model.layers.la.LogicalComponent)
    assert newobj in model.la.root_component.components


def test_created_elements_show_up_in_xml_after_adding_them(
    model: capellambse.MelodyModel,
):
    newobj = model.la.root_component.components.create(name="TestComponent")

    try:
        model._loader[newobj.uuid]
    except KeyError as err:
        raise AssertionError(
            "Cannot find added element via subscripting"
        ) from err

    assert model._loader.xpath(
        XPATH_UUID.format(newobj.uuid)
    ), "Cannot find added element via XPath"


@pytest.mark.parametrize(
    "deletion_target",
    [0, slice(None, 1)],
)
def test_deleted_elements_are_removed(
    model: capellambse.MelodyModel, deletion_target
):
    comps = model.la.root_component.components
    assert len(comps) == 2, "Precondition not met: Bad list length"

    olduuid = comps[0].uuid
    del comps[deletion_target]

    assert len(comps) != 2, "List length did not change"

    with pytest.raises(KeyError):
        model._loader[olduuid]  # pylint: disable=pointless-statement

    assert not model._loader.xpath(
        XPATH_UUID.format(olduuid)
    ), "Element is still present in tree after deleting"


def test_delete_all_deletes_matching_objects(model: capellambse.MelodyModel):
    comps = model.la.root_component.components
    assert len(comps) == 2

    comps.delete_all(name="Delete Me")
    assert len(comps) == 1
    assert comps[0].name == "Keep Me"
